@@ -1,8 +1,8 @@
-#include "pysam.h"
+#include "bcftools.pysam.h"
 
 /*  vcfplugin.c -- plugin modules for operating on VCF/BCF files.
 
-    Copyright (C) 2013-2016 Genome Research Ltd.
+    Copyright (C) 2013-2017 Genome Research Ltd.
 
     Author: Petr Danecek <pd3@sanger.ac.uk>
 
@@ -40,7 +40,11 @@ THE SOFTWARE.  */
 #include <htslib/synced_bcf_reader.h>
 #include <htslib/kseq.h>
 #include <htslib/khash_str2int.h>
+#ifdef _WIN32
+#include <windows.h>
+#else
 #include <dlfcn.h>
+#endif
 #include "bcftools.h"
 #include "vcmp.h"
 #include "filter.h"
@@ -156,7 +160,7 @@ static void add_plugin_paths(args_t *args, const char *path)
 {
     while (1)
     {
-        size_t len = strcspn(path, ":");
+        size_t len = strcspn(path, HTS_PATH_SEPARATOR_STR);
 
         if ( len == 0 )
         {
@@ -176,18 +180,18 @@ static void add_plugin_paths(args_t *args, const char *path)
                 args->plugin_paths = (char**) realloc(args->plugin_paths,sizeof(char*)*(args->nplugin_paths+1));
                 args->plugin_paths[args->nplugin_paths] = dir;
                 args->nplugin_paths++;
-                if ( args->verbose > 1 ) fprintf(pysam_stderr, "plugin directory %s .. ok\n", dir);
+                if ( args->verbose > 1 ) fprintf(bcftools_stderr, "plugin directory %s .. ok\n", dir);
             }
             else
             {
-                if ( args->verbose > 1 ) fprintf(pysam_stderr, "plugin directory %s .. %s\n", dir, strerror(errno));
+                if ( args->verbose > 1 ) fprintf(bcftools_stderr, "plugin directory %s .. %s\n", dir, strerror(errno));
                 free(dir);
             }
 
         }
 
         path += len;
-        if ( *path == ':' ) path++;
+        if ( *path == HTS_PATH_SEPARATOR_CHAR ) path++;
         else break;
     }
 }
@@ -209,28 +213,55 @@ static void *dlopen_plugin(args_t *args, const char *fname)
 
     void *handle;
     char *tmp;
-    if ( fname[0]!='/' )    // not an absolute path
+    int is_absolute_path = 0;
+#ifdef _WIN32
+    // Windows accepts both forward slash (/) and backslash (\) as folder separator
+    // and can have any path prefixed by the drive letter and a colon (:).
+    if ( fname[0]=='/' || fname[0]=='\\') is_absolute_path = 1;
+    else if ( fname[0] && fname[1]==':' && (fname[2]=='/' || fname[2]=='\\') ) is_absolute_path = 1;
+#else
+    if ( fname[0]=='/' ) is_absolute_path = 1;
+#endif
+    if ( !is_absolute_path )
     {
         int i;
         for (i=0; i<args->nplugin_paths; i++)
         {
-	    tmp = msprintf("%s/%s%s", args->plugin_paths[i], fname, PLUGIN_EXT);
+            tmp = msprintf("%s/%s%s", args->plugin_paths[i], fname, PLUGIN_EXT);
+#ifdef _WIN32
+            handle = LoadLibraryA(tmp);
+#else
             handle = dlopen(tmp, RTLD_NOW); // valgrind complains about unfreed memory, not our problem though
+#endif
             if ( args->verbose > 1 )
             {
-                if ( !handle ) fprintf(pysam_stderr,"%s:\n\tdlopen   .. %s\n", tmp,dlerror());
-                else fprintf(pysam_stderr,"%s:\n\tdlopen   .. ok\n", tmp);
+                if ( !handle )
+#ifdef _WIN32
+                    fprintf(bcftools_stderr,"%s:\n\tLoadLibraryA   .. %lu\n", tmp, GetLastError());
+#else
+                    fprintf(bcftools_stderr,"%s:\n\tdlopen   .. %s\n", tmp, dlerror());
+#endif
+                else fprintf(bcftools_stderr,"%s:\n\tplugin open   .. ok\n", tmp);
             }
             free(tmp);
             if ( handle ) return handle;
         }
     }
 
+#ifdef _WIN32
+    handle = LoadLibraryA(fname);
+#else
     handle = dlopen(fname, RTLD_NOW);
+#endif
     if ( args->verbose > 1 )
     {
-        if ( !handle ) fprintf(pysam_stderr,"%s:\n\tdlopen   .. %s\n", fname,dlerror());
-        else fprintf(pysam_stderr,"%s:\n\tdlopen   .. ok\n", fname);
+        if ( !handle )
+#ifdef _WIN32
+            fprintf(bcftools_stderr,"%s:\n\tLoadLibraryA   .. %lu\n", fname, GetLastError());
+#else
+            fprintf(bcftools_stderr,"%s:\n\tdlopen   .. %s\n", fname, dlerror());
+#endif
+        else fprintf(bcftools_stderr,"%s:\n\tplugin open   .. ok\n", fname);
     }
 
     return handle;
@@ -238,11 +269,11 @@ static void *dlopen_plugin(args_t *args, const char *fname)
 
 static void print_plugin_usage_hint(void)
 {
-    fprintf(pysam_stderr, "\nNo functional bcftools plugins were found");
+    fprintf(bcftools_stderr, "\nNo functional bcftools plugins were found");
     if ( !getenv("BCFTOOLS_PLUGINS") )
-        fprintf(pysam_stderr,". The environment variable BCFTOOLS_PLUGINS is not set.\n\n");
+        fprintf(bcftools_stderr,". The environment variable BCFTOOLS_PLUGINS is not set.\n\n");
     else
-        fprintf(pysam_stderr,
+        fprintf(bcftools_stderr,
                 " in\n\tBCFTOOLS_PLUGINS=\"%s\".\n\n"
                 "- Is the plugin path correct?\n\n"
                 "- Run \"bcftools plugin -lv\" for more detailed error output.\n"
@@ -266,25 +297,74 @@ static int load_plugin(args_t *args, const char *fname, int exit_on_error, plugi
         return -1;
     }
 
+#ifdef _WIN32
+    plugin->init = (dl_init_f) GetProcAddress(plugin->handle, "init");
+    if ( plugin->init && args->verbose > 1 ) fprintf(bcftools_stderr,"\tinit     .. ok\n");
+
+    plugin->run = (dl_run_f) GetProcAddress(plugin->handle, "run");
+    if ( plugin->run && args->verbose > 1 ) fprintf(bcftools_stderr,"\trun     .. ok\n");
+
+    if ( !plugin->init && !plugin->run )
+    {
+        if ( exit_on_error ) error("Could not initialize %s, neither run or init found \n", plugin->name);
+        else if ( args->verbose > 1 ) fprintf(bcftools_stderr,"\tinit/run .. not found\n");
+        return -1;
+    }
+
+    plugin->version = (dl_version_f) GetProcAddress(plugin->handle, "version");
+    if ( !plugin->version )
+    {
+        if ( exit_on_error ) error("Could not initialize %s: version string not found\n", plugin->name);
+        else if ( args->verbose > 1 ) fprintf(bcftools_stderr,"\tversion  .. not found\n");
+        return -1;
+    }
+
+    plugin->about = (dl_about_f) GetProcAddress(plugin->handle, "about");
+    if ( !plugin->about )
+    {
+        if ( exit_on_error ) error("Could not initialize %s: about string not found\n", plugin->name);
+        return -1;
+    }
+
+    plugin->usage = (dl_about_f) GetProcAddress(plugin->handle, "usage");
+    if ( !plugin->usage )
+        plugin->usage = plugin->about;
+
+    if ( plugin->run ) return 0;
+
+    plugin->process = (dl_process_f) GetProcAddress(plugin->handle, "process");
+    if ( !plugin->process )
+    {
+        if ( exit_on_error ) error("Could not initialize %s: process method not found\n", plugin->name);
+        return -1;
+    }
+
+    plugin->destroy = (dl_destroy_f) GetProcAddress(plugin->handle, "destroy");
+    if ( !plugin->destroy )
+    {
+        if ( exit_on_error ) error("Could not initialize %s: destroy method not found\n", plugin->name);
+        return -1;
+    }
+#else
     dlerror();
     plugin->init = (dl_init_f) dlsym(plugin->handle, "init");
     char *ret = dlerror();
     if ( ret )
         plugin->init = NULL;
     else
-        if ( args->verbose > 1 ) fprintf(pysam_stderr,"\tinit     .. ok\n");
+        if ( args->verbose > 1 ) fprintf(bcftools_stderr,"\tinit     .. ok\n");
 
     plugin->run = (dl_run_f) dlsym(plugin->handle, "run");
     ret = dlerror();
     if ( ret )
         plugin->run = NULL;
     else
-        if ( args->verbose > 1 ) fprintf(pysam_stderr,"\trun      .. ok\n");
+        if ( args->verbose > 1 ) fprintf(bcftools_stderr,"\trun      .. ok\n");
 
     if ( !plugin->init && !plugin->run )
     {
         if ( exit_on_error ) error("Could not initialize %s, neither run or init found \n", plugin->name);
-        else if ( args->verbose > 1 ) fprintf(pysam_stderr,"\tinit/run .. not found\n");
+        else if ( args->verbose > 1 ) fprintf(bcftools_stderr,"\tinit/run .. not found\n");
         return -1;
     }
 
@@ -293,7 +373,7 @@ static int load_plugin(args_t *args, const char *fname, int exit_on_error, plugi
     if ( ret )
     {
         if ( exit_on_error ) error("Could not initialize %s, version string not found\n", plugin->name);
-        else if ( args->verbose > 1 ) fprintf(pysam_stderr,"\tversion  .. not found\n");
+        else if ( args->verbose > 1 ) fprintf(bcftools_stderr,"\tversion  .. not found\n");
         return -1;
     }
 
@@ -327,6 +407,7 @@ static int load_plugin(args_t *args, const char *fname, int exit_on_error, plugi
         if ( exit_on_error ) error("Could not initialize %s: %s\n", plugin->name, ret);
         return -1;
     }
+#endif
 
     return 0;
 }
@@ -341,12 +422,12 @@ static void init_plugin(args_t *args)
     args->plugin.version(&bver, &hver);
     if ( strcmp(bver,bcftools_version()) && !warned_bcftools )
     {
-        fprintf(pysam_stderr,"WARNING: bcftools version mismatch .. bcftools at %s, the plugin \"%s\" at %s\n", bcftools_version(),args->plugin.name,bver);
+        fprintf(bcftools_stderr,"WARNING: bcftools version mismatch .. bcftools at %s, the plugin \"%s\" at %s\n", bcftools_version(),args->plugin.name,bver);
         warned_bcftools = 1;
     }
     if ( strcmp(hver,hts_version()) && !warned_htslib )
     {
-        fprintf(pysam_stderr,"WARNING: htslib version mismatch .. bcftools at %s, the plugin \"%s\" at %s\n", hts_version(),args->plugin.name,hver);
+        fprintf(bcftools_stderr,"WARNING: htslib version mismatch .. bcftools at %s, the plugin \"%s\" at %s\n", hts_version(),args->plugin.name,hver);
         warned_htslib = 1;
     }
     args->drop_header += ret;
@@ -401,11 +482,11 @@ static int list_plugins(args_t *args)
         for (i=0; i<nplugins; i++)
         {
             if ( args->verbose )
-                fprintf(pysam_stdout, "\n-- %s --\n%s", plugins[i].name, plugins[i].about());
+                fprintf(bcftools_stdout, "\n-- %s --\n%s", plugins[i].name, plugins[i].about());
             else
-                fprintf(pysam_stdout, "%s\n", plugins[i].name);
+                fprintf(bcftools_stdout, "%s\n", plugins[i].name);
         }
-        if ( args->verbose ) fprintf(pysam_stdout, "\n");
+        if ( args->verbose ) fprintf(bcftools_stdout, "\n");
     }
     else
         print_plugin_usage_hint();
@@ -429,7 +510,7 @@ static void init_data(args_t *args)
         args->out_fh = hts_open(args->output_fname,hts_bcf_wmode(args->output_type));
         if ( args->out_fh == NULL ) error("Can't write to \"%s\": %s\n", args->output_fname, strerror(errno));
         if ( args->n_threads ) hts_set_threads(args->out_fh, args->n_threads);
-        bcf_hdr_write(args->out_fh, args->hdr_out);
+        if ( bcf_hdr_write(args->out_fh, args->hdr_out)!=0 ) error("[%s] Error: cannot write to %s\n", __func__,args->output_fname);
     }
 }
 
@@ -437,7 +518,11 @@ static void destroy_data(args_t *args)
 {
     free(args->plugin.name);
     if ( args->plugin.destroy ) args->plugin.destroy();
+#ifdef _WIN32
+    FreeLibrary(args->plugin.handle);
+#else
     dlclose(args->plugin.handle);
+#endif
     if ( args->hdr_out ) bcf_hdr_destroy(args->hdr_out);
     if ( args->nplugin_paths>0 )
     {
@@ -447,34 +532,34 @@ static void destroy_data(args_t *args)
     }
     if ( args->filter )
         filter_destroy(args->filter);
-    if (args->out_fh) hts_close(args->out_fh);
+    if (args->out_fh && hts_close(args->out_fh)!=0 ) error("[%s] Error: close failed .. %s\n", __func__,args->output_fname);
 }
 
 static void usage(args_t *args)
 {
-    fprintf(pysam_stderr, "\n");
-    fprintf(pysam_stderr, "About:   Run user defined plugin\n");
-    fprintf(pysam_stderr, "Usage:   bcftools plugin <name> [OPTIONS] <file> [-- PLUGIN_OPTIONS]\n");
-    fprintf(pysam_stderr, "         bcftools +name [OPTIONS] <file>  [-- PLUGIN_OPTIONS]\n");
-    fprintf(pysam_stderr, "\n");
-    fprintf(pysam_stderr, "VCF input options:\n");
-    fprintf(pysam_stderr, "   -e, --exclude <expr>        exclude sites for which the expression is true\n");
-    fprintf(pysam_stderr, "   -i, --include <expr>        select sites for which the expression is true\n");
-    fprintf(pysam_stderr, "   -r, --regions <region>      restrict to comma-separated list of regions\n");
-    fprintf(pysam_stderr, "   -R, --regions-file <file>   restrict to regions listed in a file\n");
-    fprintf(pysam_stderr, "   -t, --targets <region>      similar to -r but streams rather than index-jumps\n");
-    fprintf(pysam_stderr, "   -T, --targets-file <file>   similar to -R but streams rather than index-jumps\n");
-    fprintf(pysam_stderr, "VCF output options:\n");
-    fprintf(pysam_stderr, "       --no-version            do not append version and command line to the header\n");
-    fprintf(pysam_stderr, "   -o, --output <file>         write output to a file [standard output]\n");
-    fprintf(pysam_stderr, "   -O, --output-type <type>    'b' compressed BCF; 'u' uncompressed BCF; 'z' compressed VCF; 'v' uncompressed VCF [v]\n");
-    fprintf(pysam_stderr, "       --threads <int>         number of extra output compression threads [0]\n");
-    fprintf(pysam_stderr, "Plugin options:\n");
-    fprintf(pysam_stderr, "   -h, --help                  list plugin's options\n");
-    fprintf(pysam_stderr, "   -l, --list-plugins          list available plugins. See BCFTOOLS_PLUGINS environment variable and man page for details\n");
-    fprintf(pysam_stderr, "   -v, --verbose               print verbose information, -vv increases verbosity\n");
-    fprintf(pysam_stderr, "   -V, --version               print version string and exit\n");
-    fprintf(pysam_stderr, "\n");
+    fprintf(bcftools_stderr, "\n");
+    fprintf(bcftools_stderr, "About:   Run user defined plugin\n");
+    fprintf(bcftools_stderr, "Usage:   bcftools plugin <name> [OPTIONS] <file> [-- PLUGIN_OPTIONS]\n");
+    fprintf(bcftools_stderr, "         bcftools +name [OPTIONS] <file>  [-- PLUGIN_OPTIONS]\n");
+    fprintf(bcftools_stderr, "\n");
+    fprintf(bcftools_stderr, "VCF input options:\n");
+    fprintf(bcftools_stderr, "   -e, --exclude <expr>        exclude sites for which the expression is true\n");
+    fprintf(bcftools_stderr, "   -i, --include <expr>        select sites for which the expression is true\n");
+    fprintf(bcftools_stderr, "   -r, --regions <region>      restrict to comma-separated list of regions\n");
+    fprintf(bcftools_stderr, "   -R, --regions-file <file>   restrict to regions listed in a file\n");
+    fprintf(bcftools_stderr, "   -t, --targets <region>      similar to -r but streams rather than index-jumps\n");
+    fprintf(bcftools_stderr, "   -T, --targets-file <file>   similar to -R but streams rather than index-jumps\n");
+    fprintf(bcftools_stderr, "VCF output options:\n");
+    fprintf(bcftools_stderr, "       --no-version            do not append version and command line to the header\n");
+    fprintf(bcftools_stderr, "   -o, --output <file>         write output to a file [standard output]\n");
+    fprintf(bcftools_stderr, "   -O, --output-type <type>    'b' compressed BCF; 'u' uncompressed BCF; 'z' compressed VCF; 'v' uncompressed VCF [v]\n");
+    fprintf(bcftools_stderr, "       --threads <int>         use multithreading with <int> worker threads [0]\n");
+    fprintf(bcftools_stderr, "Plugin options:\n");
+    fprintf(bcftools_stderr, "   -h, --help                  list plugin's options\n");
+    fprintf(bcftools_stderr, "   -l, --list-plugins          list available plugins. See BCFTOOLS_PLUGINS environment variable and man page for details\n");
+    fprintf(bcftools_stderr, "   -v, --verbose               print verbose information, -vv increases verbosity\n");
+    fprintf(bcftools_stderr, "   -V, --version               print version string and exit\n");
+    fprintf(bcftools_stderr, "\n");
     exit(1);
 }
 
@@ -584,27 +669,33 @@ int main_plugin(int argc, char *argv[])
     {
         const char *bver, *hver;
         args->plugin.version(&bver, &hver);
-        fprintf(pysam_stdout, "bcftools  %s using htslib %s\n", bcftools_version(), hts_version());
-        fprintf(pysam_stdout, "plugin at %s using htslib %s\n\n", bver, hver);
+        fprintf(bcftools_stdout, "bcftools  %s using htslib %s\n", bcftools_version(), hts_version());
+        fprintf(bcftools_stdout, "plugin at %s using htslib %s\n\n", bver, hver);
         return 0;
     }
 
     if ( usage_only )
     {
         if ( args->plugin.usage )
-            fprintf(pysam_stderr,"%s",args->plugin.usage());
+            fprintf(bcftools_stderr,"%s",args->plugin.usage());
         else
-            fprintf(pysam_stderr,"Usage: bcftools +%s [General Options] -- [Plugin Options]\n",plugin_name);
+            fprintf(bcftools_stderr,"Usage: bcftools +%s [General Options] -- [Plugin Options]\n",plugin_name);
         return 0;
     }
 
     char *fname = NULL;
     if ( optind>=argc || argv[optind][0]=='-' )
     {
-        if ( !isatty(fileno((FILE *)stdin)) ) fname = "-";  // reading from stdin
-        else usage(args);
         args->plugin.argc = argc - optind + 1;
         args->plugin.argv = argv + optind - 1;
+
+        if ( !isatty(fileno((FILE *)stdin)) ) fname = "-";  // reading from stdin
+        else if ( optind>=argc ) usage(args);
+        else
+        {
+            optind = 1;
+            init_plugin(args);
+        }
     }
     else
     {
@@ -626,7 +717,7 @@ int main_plugin(int argc, char *argv[])
             error("Failed to read the targets: %s\n", args->targets_list);
         args->files->collapse |= COLLAPSE_SOME;
     }
-    if ( !bcf_sr_add_reader(args->files, fname) ) error("Failed to open %s: %s\n", fname,bcf_sr_strerror(args->files->errnum));
+    if ( !bcf_sr_add_reader(args->files, fname) ) error("Failed to read from %s: %s\n", !strcmp("-",fname)?"standard input":fname,bcf_sr_strerror(args->files->errnum));
 
     init_data(args);
     while ( bcf_sr_next_line(args->files) )
@@ -639,7 +730,11 @@ int main_plugin(int argc, char *argv[])
             if ( !pass ) continue;
         }
         line = args->plugin.process(line);
-        if ( line ) bcf_write1(args->out_fh, args->hdr_out, line);
+        if ( line )
+        {
+            if ( line->errcode ) error("[E::main_plugin] Unchecked error (%d), exiting\n",line->errcode);
+            if ( bcf_write1(args->out_fh, args->hdr_out, line)!=0 ) error("[%s] Error: cannot write to %s\n", __func__,args->output_fname);
+        }
     }
     destroy_data(args);
     bcf_sr_destroy(args->files);
@@ -651,7 +746,7 @@ int main_plugin(int argc, char *argv[])
 
 int main_plugin(int argc, char *argv[])
 {
-    fprintf(pysam_stderr, "bcftools plugins are disabled.  To use them, you will need to rebuild\n"
+    fprintf(bcftools_stderr, "bcftools plugins are disabled.  To use them, you will need to rebuild\n"
 	    "bcftools from the source distribution with plugins enabled.\n");
     return 1;
 }
